/* -----------------------------
 * Boot Structures and Functions
 * -----------------------------
 */
#include <xen/config.h>
#include <xen/threads.h>
#include <xen/init.h>
#include <xen/linkage.h>
#include <xen/types.h>
#include <xen/smp.h>
#include <xen/spinlock.h>
#include <xen/serial.h>
#include <xen/xmalloc.h>
#include <xen/string.h>
#include <xen/serial.h>
#include <xen/keyhandler.h>
#include <xen/vga.h>
#include <xen/console.h>
#include <xen/mm.h>
#include <xen/debug.h>

#include <public/version.h>

#include <asm/cpu-info.h>
#include <asm/netl/iomap.h>
#include <asm/netl/pic.h>
#include <asm/netl/uart.h>
#include <asm/bootinfo.h>
#include <asm/netl/asmmacro.h>

#include <libfdt.h>
#include <asm/fdtmacros.h>

#define MAX_NR_ARGS 32
#define MAX_NR_ENVS 32

unsigned long fw_arg0, fw_arg1, fw_arg2, fw_arg3;
void *fdt;

struct typename dram_descriptors[] =
{
	{ 0x01, "RAM Memory" },
	{ 0x02, "ROM Memory" },
	{ 0x03, "Reserved Memory" },
	{ 0xff, "Unknown type" }
};

/* ------------------
 * Internal Functions
 * ------------------ 
 */

static char *addrtype_to_string(int type, struct typename *map, int size)
{
	int i = 0;
	int nr_elements = size/sizeof(struct typename);

	for (i = 0; i < nr_elements; i++) {
		if ((map[i].type == type ) || (map[i].type == 0xff))
			return map[i].name;
	}
	return ("Unknown type");
}

void print_addrmap(struct addrmap *addrmap, struct typename *map, int size)
{
	int i;

	if (addrmap == NULL)
		return;

	printk("(%s): nr_map = %d\n", __FUNCTION__, addrmap->nr_map);
	for (i = 0; i < addrmap->nr_map; i++) {
		printk("\t%010llx --> %010llx ( %s )\n",
			   (unsigned long long)addrmap->map[i].addr,
			   (unsigned long long)(addrmap->map[i].addr + addrmap->map[i].size - 1),
			   addrtype_to_string(addrmap->map[i].type, map, size)
			);
	}
}

extern char boot_command_line[2048];
char hypervisor_name[32];

int read_bootinfo(struct bootinfo *bootinfo)
{
	int retval __attribute__((unused));
	int intarg;
	extern int opt_dom0_hvm;
	uint32_t regs[4];

	mmio_baseaddr = MMIO_BASEADDR;
	cpu_frequency = 1600000000;

	fdt = (void *)fw_arg0;
	if (fdt == NULL)
		return -1;

	retval = fdt_check_header(fdt);
	if (retval < 0) {
		printk("FDT_ERROR: Invalid FDT Detected: %s\n", fdt_strerror(retval));
		return -1;
	}

	FDT_GET_ATTRIB_STR("/hypervisor", "hypervisor-name", hypervisor_name);

	if (get_fdt_prop("/hypervisor", "alloc_dom0_memory", &intarg, PROP_INT32) == 0)
		opt_dom0_hvm = !intarg;
	printk("opt_dom0_hvm = %d\n", opt_dom0_hvm);

	if (get_fdt_prop("/hypervisor", "bootargs", boot_command_line, PROP_STR) != 0)
		printk("failed to read boot command line args\n");
	printk("boot_command_line = \"%s\"\n", boot_command_line);

	if (get_fdt_prop("/hypervisor", "domain_heap", regs, PROP_INT32) != 0)
		printk("failed to read domain heap!\n");
	printk("domain_heap: regs[0] = %x, regs[1] = %x\n", regs[0], regs[1]);

	return 0;
}
