#include <xen/config.h>
#include <xen/init.h>
#include <xen/lib.h>
#include <xen/errno.h>
#include <xen/sched.h>
#include <xen/domain.h>
#include <xen/smp.h>
#include <xen/delay.h>
#include <xen/softirq.h>
#include <xen/grant_table.h>
#include <xen/iocap.h>
#include <xen/kernel.h>
#include <xen/multicall.h>
#include <xen/irq.h>
#include <xen/event.h>
#include <xen/console.h>
#include <xen/percpu.h>
#include <xen/debug.h>
#include <xen/version.h>
#include <xen/libelf.h>
#include <xen/shared.h>
#include <xen/mm.h>
#include <xen/xmalloc.h>
#include <xen/sched-if.h>
#include <public/xen.h>

#include <asm/cpu.h>
#include <asm/processor.h>
#include <asm/pgalloc.h>
#include <asm/thread_info.h>
#include <asm/system.h>
#include <asm/mapops.h>
#include <asm/pgwalker.h>
#include <asm/traps.h>

// #define VERBOSE

#define align(addr, size) ((addr + size - 1) & ~(size - 1))
#define PTR_ALIGN(val) align(val, sizeof(void *))

unsigned long dom0_size;
unsigned long dom0_memory_start, dom0_memory_size;
unsigned long dom0_initrd_start, dom0_initrd_size;

unsigned long (*get_free_pages)(gfp_t gfp_mask, unsigned int order);

extern unsigned long xensp[NR_CPUS];

/*
 * struct vcpu is akin to "task_struct" in Linux,
 * and capture a process/vcpu as the case may be
 *
 * 'struct vcpu' is created for every vcpu of each
 * guest domain. $28 on each cpu could point to
 * 'struct vcpu' structure corresponding to the
 * vcpu instance of the current guest domain
 *
 * the schedule() function that is eventually called
 * in context_switch() is the entry point into
 * guest domain's schedule function and must be
 * populated in the arch-speci component of
 * struct vcpu
 */

/*
 * I guess schedule_tail is called for the first run of a guest domain instance
 * just as schedule_tail() is called for the child process from ret_from_fork()
 */

pte_t null_pte = {(_CACHE_CACHABLE_COW | _PAGE_DIRTY |_PAGE_GLOBAL) >> 6}; /* _PAGE_VALID is off */

void schedule_tail(struct vcpu *prev)
{
#if 0
	extern char ia64_ivt;

	context_saved(prev); /* complete the context switch of the prev guest domain */

	
	load_state(current); /* load registers from the current thread state and off it goes */
	// migrate_timer(&current->arch.hlt_timer, current->processor);
	flush_vtlb_for_context_switch(prev, current);
#endif
}

#ifndef CONFIG_XEN_PGWALKER

#define PTE_PFN_SHIFT 6

void setup_tlbs(struct domain *d)
{
	pte_t *even_ptep, *odd_ptep;
	unsigned long addr;
	int tlb_index = 0, nr_pages;
	int i = 0;

	while (i < d->arch.nr_map_entries) {
		nr_pages = 0;

#ifdef DEBUG
		printk("setting the tlbs vaddr=0x%lx, pages=%ld\n",
			   d->arch.map_regions[i].virt_start,
			   d->arch.map_regions[i].nr_pages);
#endif

		for (addr = d->arch.map_regions[i].virt_start;
			 nr_pages < d->arch.map_regions[i].nr_pages;
			 nr_pages += 2, addr += 2 * d->arch.page_size) {

			/* read the odd even pair. worry about optimizations later */
			even_ptep = lookup_domain_pte(d, addr);
			if (even_ptep == NULL)
				even_ptep = &null_pte;
			odd_ptep = lookup_domain_pte(d, addr + d->arch.page_size);
			if (odd_ptep == NULL)
				odd_ptep = &null_pte;
			write_c0_entryhi(addr);
			write_c0_entrylo0(pte_val(*even_ptep) >> PTE_PFN_SHIFT);
			write_c0_entrylo1(pte_val(*odd_ptep) >> PTE_PFN_SHIFT);
			/*
			 * FIXME: replace DOM_TLB_PAGEMASK with a variable pagemask
			 *        that can be populated in the architecture-specific
			 *        domain information structure "struct arch_domain"
			 */
			write_c0_pagemask(DOM_TLB_PAGEMASK);

			write_c0_index(tlb_index++);
			tlb_write_indexed();
		}
		i++;
	}
}

#endif

void context_switch(struct vcpu *prev, struct vcpu *next)
{
	struct domain *d __attribute__((unused)) = next->domain;

	/*
	 * TODO: TLB needs to be flushed on context switches
	 *       as we do not yet have the concept of domain
	 *       id, similar to ASID
	 *       
	 *       TLB need not be flushed in its entirety if
	 *       vcpu (or task) being scheduled belongs to the same 
	 *       domain as the prev task
	 *
	 *       If hypervisor itself has mapped entries,
	 *       could we avoid flushing entries belonging to
	 *       Hypervisor ?
	 *
	 *       In linux world, the mmu context is switched here
	 */

	/*
	 * In resume(), registers are setup, i.e., the next context
	 * is switched in. TLBs need to be wired/setup prior to
	 * calling resume()
     */

	local_irq_enable();

	if (!is_hvm_domain(d))	{
#ifdef CONFIG_XEN_PGWALKER
		setup_user_pgd(d->arch.mm.pgd);
#else
		setup_tlbs(d);
#endif
	}
	
	context_saved(prev);
	resume(prev, next, task_thread_info(next));
	// switch_to(prev, next);
}

void vcpu_kick(struct vcpu *v)
{

}

void vcpu_init_regs(struct vcpu *v)
{
	struct vcpu_guest_context *regs = &(v->arch.guest_context);
	
	regs->reg16 = 0x0;
}

/*
 * architecture-specific API
 */

struct vcpu *alloc_vcpu_struct(void)
{
	/*
	 * called in common/domain.c:alloc_vcpu()
	 */
    struct vcpu *v;
	static int first_allocation = 1;
	struct thread_info *thread_info;

	if (first_allocation) {
		first_allocation = 0;
		/* Still keep idle vcpu0 static allocated at compilation, due
		 * to some code from Linux still requires it in early phase.
		 */
		return idle_vcpu[0];
	}

	/*
	 * QUESTION: Where does xmalloc() gets its memory from ?
	 */
	v = xmalloc(struct vcpu);
    if (v == NULL) {
		dbg();
		return NULL;
	}

	memset(v, 0, sizeof(*v));

	/*
	 *  FIXME: alloc_thread_info() macro is partially defined in
	 *         include/asm/thread_info.h. Hide the following call
	 *         to alloc_xenheap_pages() in that macros and use it
	 */
    thread_info = (struct thread_info *) alloc_xenheap_pages(THREAD_SIZE_ORDER, MEMF_bits(32));
	if (thread_info == NULL) {
		dbg();
		xfree(v);
		v = NULL;
		goto exit;
	}
	v->arch.thread_info = thread_info;
	thread_info->task = v;
	
exit:
    return v;
}

int arch_set_info_guest(struct vcpu *v, vcpu_guest_context_u c)
{
	vcpu_guest_context_t *ctxt = &v->arch.guest_context;

	memcpy(ctxt, c.nat, sizeof(vcpu_guest_context_t));
	(v->arch.guest_context).reg4 = (unsigned long)fdt;
	(v->arch.guest_context).reg5 = (unsigned long)fw_arg1;
	(v->arch.guest_context).reg6 = (unsigned long)v->domain->domain_id;

	/*
	 * XXX When do we bring rest of the vcpus online?
	 */
	if (v->is_initialised == 1)
	  return 0;

	clear_bit(_VPF_down, &v->pause_flags);

	v->is_initialised = 1;

	return 0;
}

void arch_get_info_guest(struct vcpu *v, vcpu_guest_context_u c)
{

}

int arch_domain_create(struct domain *d, unsigned int domcr_flags)
{
	int i;

	d->shared_info = alloc_xenheap_pages(get_order_from_bytes(XSI_SIZE), 0);
	if (d->shared_info == NULL)
		goto fail_nomem;

	/*
	 * This page has to be made shareable with guest domains
	 * In mips we should just pass the page frame number of
	 * the shared page to guest domains and allow guest domains
	 * to map the physical page to a convenient virutal address
	 */
	memset(d->shared_info, 0, XSI_SIZE);

	if ((d->arch.mm.pgd = pgd_alloc(&d->arch.mm)) == NULL)
	    goto fail_nomem;

	if (domcr_flags & DOMCRF_hvm)
		d->arch.page_size = PAGE_SIZE;
	else
		d->arch.page_size = DOM_PAGE_SIZE;

	d->arch.relmem = RELMEM_not_started;
	INIT_PAGE_LIST_HEAD(&d->arch.relmem_list);

	for (i = 0; i < NR_CPUS; ++i)
		d->shared_info->arch.xensp[i] = &xensp[i];

	d->shared_info->arch.dom_page_shift = DOM_PAGE_SHIFT;
#ifdef CONFIG_XEN_2LEVEL_PGTABLE
	d->shared_info->arch.pgtable_levels = 2;
#else
	d->shared_info->arch.pgtable_levels = 3;
#endif

	return 0;

fail_nomem:
	if (d->arch.mm.pgd != NULL)
	    pgd_free(d->arch.mm.pgd);
	if (d->shared_info != NULL)
	    free_xenheap_pages(virt_to_page(d->shared_info), get_order_from_bytes(XSI_SIZE));
	return -ENOMEM;
}

void arch_domain_destroy(struct domain *d)
{
	if (d->shared_info != NULL) {
		free_xenheap_pages(d->shared_info,
						   get_order_from_bytes(XSI_SIZE));
	}
}

void free_domain_struct(struct domain *d)
{
    free_xenheap_pages(d, get_order_from_bytes(sizeof(*d)));
}

void vcpu_destroy(struct vcpu *v)
{

}

static int relinquish_memory(struct domain *d, struct page_list_head *list)
{
	struct page_info *page;
	int ret = 0;

	/* Use a recursive lock, as we may enter 'free_domheap_page'. */
	spin_lock_recursive(&d->page_alloc_lock);

	while ( (page = page_list_remove_head(list)) ) {
		
		/* Grab a reference to the page so it won't disappear from under us. */
        if ( unlikely(!get_page(page, d)) )
        {
            /* Couldn't get a reference -- someone is freeing this page. */
            page_list_add_tail(page, &d->arch.relmem_list);
            continue;
        }
		/*
		 * XXX Need to test the behaviour in case of grant table pages
		 * where multiple domains may be holding references to a page.
		 */
        if ( test_and_clear_bit(_PGC_allocated, &page->count_info) ) {
            put_page(page);
		}
		
		page_list_add_tail(page, &d->arch.relmem_list);

        put_page(page);
	}

    /* list is empty at this point. */
    if ( !page_list_empty(&d->arch.relmem_list) )
    {
        *list = d->arch.relmem_list;
        INIT_PAGE_LIST_HEAD(&d->arch.relmem_list);
    }

    spin_unlock_recursive(&d->page_alloc_lock);
    return ret;
}

int domain_relinquish_resources(struct domain *d)
{
	int ret = 0;

	switch(d->arch.relmem) {
	case RELMEM_not_started:
		/* Tear down the page tables */
		mm_teardown(d);
		d->arch.relmem = RELMEM_dom;
		/* Nothing to do here - fallthrough */

	case RELMEM_dom:
		/* Relinquish every domain page of memory. */
		ret = relinquish_memory(d, &d->page_list);
		if (ret != 0)
			return ret;
		d->arch.relmem = RELMEM_xen;
		/* fallthrough */

	case RELMEM_xen:
		/* Relinquish every xen page of memory. */
		ret = relinquish_memory(d, &d->xenpage_list);
		if (ret != 0)
			return ret;
		
		d->arch.relmem = RELMEM_done;
		/* fallthrough */

	case RELMEM_done:
		break;

	default:
		BUG();
	}

	return ret;
}

void arch_vcpu_reset(struct vcpu *v)
{
	
}

void vcpu_mark_events_pending(struct vcpu *v)
{
	
}

void free_vcpu_struct(struct vcpu *v)
{
	if (v->arch.thread_info != NULL) {
		free_xenheap_pages(v->arch.thread_info, THREAD_SIZE_ORDER);
	}
	xfree(v);
}

int vcpu_initialise(struct vcpu *v)
{
	return 0;
}

struct domain *alloc_domain_struct(void)
{
    struct domain *d;
    /*
     * We pack the MFN of the domain structure into a 32-bit field within
     * the page_info structure. Hence the MEMF_bits() restriction.
	 *
	 * FIXME: Are we better-off using xmalloc() and not waste a whole
	 *        page for domain structure ?
     */
    d = alloc_xenheap_pages(get_order_from_bytes(sizeof(*d)), MEMF_bits(32 + PAGE_SHIFT));
    if (d != NULL)
        memset(d, 0, sizeof(*d));
	if (d == NULL) {
		dbg();
		cpu_halt();
	}
    return d;
}

long arch_do_vcpu_op(int cmd, struct vcpu *v, XEN_GUEST_HANDLE(void) arg)
{
	return 0;
}

void arch_dump_domain_info(struct domain *d)
{

}

void arch_dump_vcpu_info(struct vcpu *v)
{

}

void continue_running(struct vcpu *same)
{

}

void sync_vcpu_execstate(struct vcpu *v)
{

}

void dump_pageframe_info(struct domain *d)
{   
    // printk("dump_pageframe_info not implemented\n");
}

/*
 * mpaddr = metaphysical address (physical address)
 *  paddr = machine address
 */

void loaddomainelfimage(struct domain *d, struct elf_binary *elf,
							   unsigned long phys_load_offset)
{
	const elf_phdr *phdr;
	int phnum, h, filesz, memsz;
	unsigned long elfaddr, dom_mpaddr, dom_imva;
	struct page_info *p;
	unsigned int page_size = d->arch.page_size;

	phnum = elf_uval(elf, elf->ehdr, e_phnum);
	for (h = 0; h < phnum; h++) {
		phdr = elf_phdr_by_index(elf, h);
		if (!elf_phdr_is_loadable(elf, phdr))
		    continue;

		filesz = elf_uval(elf, phdr, p_filesz);
		memsz = elf_uval(elf, phdr, p_memsz);

		elfaddr = (unsigned long) elf->image + elf_uval(elf, phdr, p_offset);
		dom_mpaddr = elf_uval(elf, phdr, p_paddr);
		dom_mpaddr += phys_load_offset;

#ifdef DEBUG
		printk("seg(%d): dom_mpaddr = 0x%lx, filesz = 0x%x, memsz = 0x%x\n",
			   h, dom_mpaddr, filesz, memsz);
#endif

		/*
		 * XXX: dom_mpaddr represents the virtual address of launched
		 *      domains. It cannot be masked with 0x7fffffff as the
		 *      address space could be larger than 31 bits. Use the
		 *      the maximum allowed number of virtual adddress bits
		 *      Of course, the address space will have to be accommodated
		 *      by page tables (be it 2-level or 3-level)
		 */
		if (is_hvm_domain(d))
			dom_mpaddr &= 0x3fffffffULL;
		else
			dom_mpaddr &= 0xffffffffffUL; /* 40-bit vaddr space */
				
		while (memsz > 0) {
			if (is_hvm_domain(d))
				/*
				 * XXX: the trick of translating Linux (is_hvm == 1) physical
				 *      address into a page strcuture is ok as long as we do
				 *      not acess the page structure itself and content with
				 *      arithmetic. There are only as many pages as the number
				 *      of physical frames (contiguous) that Xen knows about
				 *      Since Linux's physical memory is out of view and bounds
				 *      of Xen (till we paravirutlaize domains), page accesses
				 *      corresponding to Linux's memory are a "no-no"
				 */
				p = maddr_to_page(dom_mpaddr);
			else
				p = assign_new_domain_page(d, dom_mpaddr);
			BUG_ON (unlikely(p == NULL));
			dom_imva = (unsigned long)__va(page_to_maddr(p));
#ifdef DEBUG
			printk("maddr == 0x%lx\n", page_to_maddr(p));
#endif
			if (filesz > 0) {
				if (filesz >= page_size) {
					copy_page((void *) dom_imva, (void *) elfaddr);
				}
				else {
					memcpy((void *) dom_imva, (void *) elfaddr, filesz);
					memset((void *) dom_imva + filesz, 0, page_size - filesz);
				}
				if (elf_uval(elf, phdr, p_flags) & PF_X) {
					flush_icache_range(dom_imva, dom_imva + page_size);
				}
			}
			else if (memsz > 0) {
				/* always zero out entire page */
				memset((void *) dom_imva, 0, page_size);
			}
			memsz -= page_size;
			filesz -= page_size;
			elfaddr += page_size;
			dom_mpaddr += page_size;
		}
	}
}

#define play_dead cpu_halt

static void continue_cpu_idle_loop(void)
{
	int cpu = smp_processor_id();

	for (;;) {
	    // page_scrub_schedule_work();
	    while (!softirq_pending(cpu))  
			play_dead();

	    raise_softirq(SCHEDULE_SOFTIRQ);
	    do_softirq();
	    if (!cpu_online(cpu))
	        play_dead();
	}
}

void startup_cpu_idle_loop(void)
{
	/* Just some sanity to ensure that the scheduler is set up okay. */
	ASSERT(current->domain->domain_id == IDLE_DOMAIN_ID);
	raise_softirq(SCHEDULE_SOFTIRQ);
	continue_cpu_idle_loop();
}


extern unsigned long  xen_pstart;

static void calc_dom0_size(void)
{
#if 0
	unsigned long domheap_pages;
	unsigned long p2m_pages;
	unsigned long spare_hv_pages;
	unsigned long max_dom0_size;
	unsigned long iommu_pg_table_pages = 0;

	/* Estimate maximum memory we can safely allocate for dom0
	 * by subtracting the p2m table allocation and a chunk of memory
	 * for DMA and PCI mapping from the available domheap pages. The
	 * chunk for DMA, PCI, etc., is a guestimate, as xen doesn't seem
	 * to have a good idea of what those requirements might be ahead
	 * of time, calculated at 128MB + 1MB per 4GB of system memory */
	domheap_pages = avail_domheap_pages();
	p2m_pages = domheap_pages / PTRS_PER_PTE;
	spare_hv_pages = 8192 + (domheap_pages / 4096);

	if (iommu_enabled)
		iommu_pg_table_pages = domheap_pages * 4 / 512;
	/* There are 512 ptes in one 4K vtd page. */

	max_dom0_size = (domheap_pages - (p2m_pages + spare_hv_pages) -
					 iommu_pg_table_pages) * PAGE_SIZE;
	printk("Maximum permitted dom0 size: %luMB\n",
	       max_dom0_size / (1024*1024));

	/* validate proposed dom0_size, fix up as needed */
	if (dom0_size > max_dom0_size) {
		printk("Reducing dom0 memory allocation from %luK to %luK "
		       "to fit available memory\n",
		       dom0_size / 1024, max_dom0_size / 1024);
		dom0_size = max_dom0_size;
	}

	/* dom0_mem=0 can be passed in to give all available mem to dom0 */
	if (dom0_size == 0) {
		printk("Allocating all available memory to dom0\n");
		dom0_size = max_dom0_size;
	}

	/* Check dom0 size.  */
	if (dom0_size < 4 * 1024 * 1024) {
		panic("dom0_mem is too small, boot aborted"
			  " (try e.g. dom0_mem=256M or dom0_mem=65536K)\n");
	}
#endif

	/* dom0_size is set while parsing FDT */
	if (!dom0_size)
		dom0_size = 128*1024*1024;

	/* no need to allocate pages for now
	 * pages are allocated by map_new_domain_page() via loaddomainelfimage()
	 */
}

#define PTR_SIZE (sizeof(void *))

extern unsigned long hypercall_table[];
/*
 * ensure that addresses passed to and operated in 
 * the following function are virutal addresses.
 */
int construct_dom0(struct domain *d,
				   unsigned long image_start, unsigned long image_len,
				   unsigned long initrd_start, unsigned long initrd_len,
				   char *cmdline)
{
	int rc;
	int i __attribute__((unused));
	struct vcpu *v  __attribute__((unused)) = d->vcpu[0];
	unsigned long max_pages;

	struct elf_binary elf;
	struct elf_dom_parms parms;
	unsigned long p_start;
	unsigned long pkern_start;
	start_info_t *si;
	struct page_info *start_info_page;

	struct page_info *hypercall_table_page;
	unsigned char *virt_hypercall;

	unsigned long pkern_entry; /* entry address of guest domains */
	unsigned long pkern_end; /* end paddr of guest domains */
	unsigned long dst;
	int offset;

	/*
	 * pstart_info is the marker for domain 0's paddr
	 * areas. It is initialized to the end of Xen paddr
	 * and extended as physical memory is allocated to
	 * such things as the ramdisk etc. 

	 * FIXME: In IA-64 mem for ramdisk seem to be allocated 
	 * before the dom0 image ; this might have to be changed 
	 * for mips
	 */
	unsigned long  pstart_info;
//	unsigned long phys_load_offset;
	int nr_entries;

	printk("*** LOADING DOMAIN 0 ***\n");

	/*
	 * FIXME: calculate the size of the domain
	 */
	calc_dom0_size();
	max_pages = dom0_size / PAGE_SIZE;
	d->max_pages = max_pages;

	/*
	 * To begin with tot_pages must be set to zero for domains
	 * whose memory is allocted by Xen (from domain heap). Else
	 * the check (tot_pages >= max_pages) becomes true and no
	 * more memory can be allocated for domains which is clearly
	 * flawed. Setting tot_pages to zero to begin with displays
	 * incorrect information for domains (such as Liunx) whose
	 * memory is not allocated by Xen. So, we make distinction
	 * through the following if condition for the present. It
	 * should go away in a truly virtualized implementation when
	 * memory for all domains is allocated by Xen's page allocator
	 */
	if (is_hvm_domain(d))
		d->tot_pages = max_pages;
	else
		d->tot_pages = 0;

	rc = elf_init(&elf, (void*)image_start, image_len);
	if (rc != 0) {
		printk("Invalid elf binary\n");
	    return rc;
	}
#ifdef VERBOSE
	elf_set_verbose(&elf);
#endif
	elf_parse_binary(&elf);
	rc = elf_xen_parse(&elf, &parms);
	if (rc != 0)
		return rc;

	/*
	 * We cannot rely on the load address in the ELF headers to
	 * determine the meta physical address at which the image
	 * is loaded.  Patch the address to match the real one, based
	 * on xen_pstart
	 */
	printk(" Dom0 kernel: %s, %s, paddr 0x%" PRIx64 " -> 0x%" PRIx64 "\n",
	       elf_64bit(&elf) ? "64-bit" : "32-bit",
	       elf_msb(&elf)   ? "msb"    : "lsb",
	       elf.pstart, elf.pend);

	if (!elf_64bit(&elf) || elf_uval(&elf, elf.ehdr, e_machine) != EM_MIPS) {
		printk("Incompatible kernel binary\n");
		return -1;
	}

	p_start = parms.virt_base;	
	if (is_hvm_domain(d)) {
		pkern_start = parms.virt_kstart & 0x3fffffffULL;
		pkern_end = parms.virt_kend & 0x3fffffffULL;
	}
	else {
		pkern_start = parms.virt_kstart;
		pkern_end = parms.virt_kend;
	}
	pkern_entry = parms.virt_entry;

	if ((p_start & (PAGE_SIZE-1)) != 0) {
	    printk("Initial guest OS must load to a page boundary.\n");
	    return -EINVAL;
	}

	/* 
	 * Copy the OS image.
	 */
//	loaddomainelfimage(d, &elf, pkern_start);
	loaddomainelfimage(d, &elf, 0);

	/* Setup hypercall page inside kernel code segment */

	pstart_info = PAGE_ALIGN(pkern_end);
	if (is_hvm_domain(d)) {
		start_info_page = virt_to_page(__va(pstart_info));
		hypercall_table_page = virt_to_page(__va(parms.virt_hypercall & 0x3fffffff));
	}
	else {
		start_info_page = assign_new_domain_page(d, pstart_info);
		hypercall_table_page = assign_new_domain_page(d, parms.virt_hypercall);
	}
	if (start_info_page == NULL || hypercall_table_page == NULL)
		panic("can't allocate start info page/hypercall_table_page");

	virt_hypercall = page_to_virt(hypercall_table_page);
	offset = parms.virt_hypercall & (d->arch.page_size - 1);
	memcpy(virt_hypercall + offset, hypercall_table, PAGE_SIZE);

	offset = pstart_info & (d->arch.page_size - 1);
	dst = (unsigned long)page_to_virt(start_info_page) + offset;

#ifdef DEBUG
	printk("paddr(start_info page) = 0x%lx\n", __pa(dst));
	printk("sizeof(bootinfo) = %ld sizeof(addrmap) = %ld\n", 
		   sizeof(bootinfo), sizeof(struct addrmap));
#endif

	si = (start_info_t *)(dst);
	clear_page(si);
	snprintf(si->magic, sizeof(si->magic), "xen-3.0-mips");
	si->nr_pages     = max_pages;
	si->flags = SIF_INITDOMAIN | SIF_PRIVILEGED;
	si->flags |= (xen_processor_pmbits << 8) & SIF_PM_MASK;
	si->pt_base = (unsigned long) d->arch.mm.pgd;
	si->mfn_list = (unsigned long)machine_to_phys_mapping;
	/*
	 * FIXME: Xen comments describe the field "shared_info"
	 *        of start page as "MACHINE address of shared 
	 *        info struct". In the following the field is
	 *        set to the virtual address (of Xen address space)
	 *        to get going and since everyone runs in kernel
	 *        mode for the present
	 */
	si->shared_info = (unsigned long) d->shared_info;
	dst += PTR_ALIGN(sizeof(start_info_t));

	nr_entries = d->arch.nr_map_entries;
	d->arch.map_regions[nr_entries].virt_start = pkern_start;
	d->arch.map_regions[nr_entries].nr_pages =
		(align(pstart_info + PAGE_SIZE, d->arch.page_size) - pkern_start)
		/ PAGE_SIZE;
	d->arch.nr_map_entries++;

	d->shared_info->arch.start_info_pfn = pstart_info >> PAGE_SHIFT;
	d->shared_info->arch.dom_page_shift = DOM_PAGE_SHIFT;
#ifdef CONFIG_XEN_2LEVEL_PGTABLE
	d->shared_info->arch.pgtable_levels = 2;
#else
	d->shared_info->arch.pgtable_levels = 3;
#endif

	v->is_initialised = 1;
	clear_bit(_VPF_down, &v->pause_flags);
	d->is_paused_by_controller = 0;

	(v->arch.guest_context).reg4 = (unsigned long)fdt;
	(v->arch.guest_context).reg5 = (unsigned long)fw_arg1;
	(v->arch.guest_context).reg6 = (unsigned long)0;
	(v->arch.guest_context).reg31 = pkern_entry;

	return 0;
}

/*
 * Temporary hypercall for destroying a guest domain -
 * Whenever the guest gets a IRQ_IPI_SMP_FUNCTION, it assumes a destroy
 * operation and relinquishes control by calling this hypercall.
 */
long do_destroy_domain(void)
{
	/* Restore xen trap handlers */
	local_irq_disable();
	trap_init();
	local_irq_enable();

	/* Flush the local tlb entries used by the guest */
	local_flush_tlb_all();

	/*
	 * The softirq handler should call schedule() to take the current
	 * vcpu offline.
	 */
	do_softirq();

	return 0;
}
