#include <xen/config.h>
#include <xen/types.h>
#include <xen/lib.h>
#include <xen/mm.h>
#include <xen/guest_access.h>
#include <xen/compat.h>
// #include <xen/pci.h>
#include <public/domctl.h>
#include <xen/sched.h>
#include <xen/domain.h>
#include <xen/event.h>
#include <xen/domain_page.h>
// #include <asm/msr.h>
#include <xen/trace.h>
#include <xen/console.h>
#include <xen/iocap.h>
// #include <xen/paging.h>
#include <asm/irq.h>
#include <asm/processor.h>
#include <xsm/xsm.h>
#include <xen/iommu.h>

#include <public/arch-mips/domctl.h>

extern unsigned long hypercall_table[];

/*
 * Setup the bootstrap page tables for the guest domain.
 */
long setup_guest_page_tables(struct domain *d,
				    xen_domctl_arch_setup_t *arch_setup)
{
	start_info_t *si;
	unsigned long mfn, pfn;
	long ret = 0;
	int i = 0, index;
	pte_t *pte;

	for (i = 0; i < arch_setup->nr_map_entries; i++) {
		/*
		 * Copy the domain virtual map regions.
		 */
		index = d->arch.nr_map_entries;
		d->arch.map_regions[index].virt_start = arch_setup->map_entries[i].map_addr;
		d->arch.map_regions[index].nr_pages = arch_setup->map_entries[i].nr_pages;
		d->arch.nr_map_entries++;
	}
	
	/*
	 * The last virtual region corresponds to start_info_t. Setup the 
	 * pgd in the start info. Rest of the start_info is populated by
	 * the user space.
	 */
	pfn = arch_setup->map_entries[d->arch.nr_map_entries - 1].pfn_start;

	pte = lookup_domain_pte(d, pfn << PAGE_SHIFT);
	if (unlikely(pte_none(*pte)))
		return -EFAULT;

	mfn = pte_val(*pte) >> PAGE_SHIFT;
	if (unlikely(!mfn_valid(mfn)))
		return -EFAULT;

	si = (start_info_t *)mfn_to_virt(mfn);
	clear_page(si);

	si->pt_base = (unsigned long)d->arch.mm.pgd;
	si->mfn_list = (unsigned long)machine_to_phys_mapping;

	return ret;
}

/*
 * Public API
 */

long arch_do_domctl(struct xen_domctl *domctl,
					XEN_GUEST_HANDLE(xen_domctl_t) u_domctl)
{
	long ret = 0;

	switch (domctl->cmd) {
		
	case XEN_DOMCTL_arch_setup:
		/*
		 * This hypercall is done by the userspace to setup
		 * the page tables while creating a domain.
		 */
		{
			struct domain *d;
			ret = -ESRCH;
			d = rcu_lock_domain_by_id(domctl->domain);
			if ( d == NULL )
				break;
			
			ret = setup_guest_page_tables(d, &domctl->u.arch_setup);

			rcu_unlock_domain(d);
		}
		break;
		
	case XEN_DOMCTL_hypercall_init:
		{
			struct domain *d = rcu_lock_domain_by_id(domctl->domain);
			unsigned long gmfn = domctl->u.hypercall_init.gmfn;
			unsigned long mfn;
			void *hypercall_page;

			ret = -ESRCH;
			if ( unlikely(d == NULL) )
				break;
			
			mfn = gmfn;
			ret = -EACCES;
			if (!mfn_valid(mfn)) {
				rcu_unlock_domain(d);
				break;
			}
			
			/*
			 * Copy the hypercall table for the domain.
			 */
			hypercall_page = mfn_to_virt(mfn);
			memcpy(hypercall_page, hypercall_table, PAGE_SIZE);
			ret = 0;

			rcu_unlock_domain(d);
		}
		break;

	default:
		ret = -ENOSYS;
		break;
	}
	
	return ret;
}

long do_domctl2(XEN_GUEST_HANDLE(xen_domctl_t) u_domctl)
{
    long ret = 0;
    struct xen_domctl curop, *domctl = &curop;

    if (copy_from_guest(domctl, u_domctl, 1))
        return -EFAULT;

    if (domctl->interface_version != XEN_DOMCTL_INTERFACE_VERSION)
        return -EACCES;

	switch (domctl->cmd) {

	case XEN_DOMCTL2_destroydomain:
		do_destroy_domain();
		break;

	case XEN_DOMCTL2_bootvcpu: {
		struct domain *d;
		vcpu_guest_context_u c = { .nat = NULL };
		unsigned int vcpu;
		struct vcpu *v;

		ret = -ESRCH;
		d = rcu_lock_domain_by_id(domctl->domain);
		if (d == NULL)
			goto bvc_out;

		ret = -EINVAL;
		if (current->domain != d)
			goto bvc_out;

		vcpu = domctl->u.vcpucontext.vcpu;
		if ((vcpu >= MAX_VIRT_CPUS) || ((v = d->vcpu[vcpu]) == NULL))
			goto bvc_out;

		if (guest_handle_is_null(domctl->u.vcpucontext.ctxt)) {
			vcpu_reset(v);
			ret = 0;
			goto bvc_out;
		}

		ret = -ENOMEM;
		if ((c.nat = xmalloc(struct vcpu_guest_context)) == NULL)
			goto bvc_out;
				
		ret = copy_from_guest(c.nat, domctl->u.vcpucontext.ctxt, 1);
		ret = ret ? -EFAULT : 0;
		if (ret == 0) {
			vcpu_pause(v);
			ret = arch_set_info_guest(v, c);
			vcpu_unpause(v);
		}

		xfree(c.nat);

	bvc_out:
		rcu_unlock_domain(d);
	}
	break;
    
	default:
		ret = -ENOSYS;
		break;
	}
	
	return ret;
}
