#include <xen/config.h>

#include <public/xen.h>
#include <public/memory.h>
#include <public/arch-mips/memory.h>

#include <xen/init.h>
#include <xen/sched.h>
#include <xen/domain.h>
#include <xen/lib.h>
#include <xen/mm.h>
#include <xen/paging.h>
#include <xen/guest_access.h>
#include <xen/debug.h>

#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/pgalloc.h>
#include <asm/pgwalker.h>

// struct page_info *frame_table __read_mostly;

struct page_info *frame_table;
unsigned long max_page;

pgd_t invalid_pg_dir[PTRS_PER_PGD];
pte_t invalid_pte_table[PTRS_PER_PTE];
pmd_t invalid_pmd_table[PTRS_PER_PMD];

unsigned long xen_pstart;

extern char _text[];

/*
 * Misc Notes
 * ----------
 *
 * VHPT in IA-64 refers to "Virtually Hashed Page Table". It is
 * equivalent to the hierarchical (tree) page tables in MIPS and 
 * x86.
 * 
 * Page tables are invalidated only after the TLB is flushed ?
 * There will still be a race condition (a livelock rather), 
 * no matter how many times tlbs are flushed before invalidating
 * the page table. However, we can disable the interrupts
 * before flushing the tlb
 *
 * Some notes on Xen's page frames:
 *
 * Xen pages are used for different purposes and the pages are
 * marked accordingly. The bits are captured in a LONG field.
 * The semantics of pages differs with the use of the page. I
 * guess "semantics" refer to the semantics of allocation and
 * deallocation of pages through the API:
 *
 * put_page(struct page_info *page)
 * get_page(struct page_info *page)
 *
 * ------------------------------------------------------------
 * | 3 bits |
 * ------------------------------------------------------------
 *
 * 0: PGT_none
 * 1: PGT_l1_page_table
 * 2: PGT_l2_page_table
 * 3: PGT_l3_page_table
 * 4: PGT_l4_page_table
 * 5: Reserved
 * 6: Reserved
 * 7: PGT_writable_page
 */

unsigned long _page_cachable_default = (0x3 << 9);

pgprot_t protection_map[16];

#define arch_vm_get_page_prot(vm_flags) __pgprot(_PAGE_DIRTY | _PAGE_VALID | _PAGE_GLOBAL)

pgprot_t vm_get_page_prot(unsigned long vm_flags)
{
	return __pgprot(pgprot_val(protection_map[vm_flags &
				(VM_READ|VM_WRITE|VM_EXEC|VM_OWNER)]) |
			pgprot_val(arch_vm_get_page_prot(vm_flags)));
}

static void *page_address(struct page_info *page)
{
	return lowmem_page_address(page);
}

/*
 * Public API
 */

struct domain *dom_xen, *dom_io;

void put_page_type(struct page_info *page)
{

}

int guest_physmap_mark_populate_on_demand(struct domain *d, 
										  unsigned long gfn, 
										  unsigned int order)
{
    return -1;
}


unsigned long gmfn_to_mfn(struct domain *d, 
								  unsigned long gpfn)
{
	return 0;
}

int get_page_type(struct page_info *page, unsigned long type)
{
	return 0;
}

static void __domain_put_page(struct domain* d, unsigned long gpaddr,
							  volatile pte_t* ptep, pte_t old_pte,
							  struct page_info* page)
{
	/*
	 * Should shadow paging be enabled, books for shadow paging
	 * will have to be maintained here, something on the lines 
	 * of IA-64's
	 *
	 * if (shadow_mode_enabled(d))
	 * shadow_mark_page_dirty(d, gpaddr >> PAGE_SHIFT);
	 */
        
	/*
	 * TODO: For virtualization, in the absence of DOM ids
	 *       in TLB, the entire tlb might have to be flushed
	 *       For now, flush tlbs for just the page address range
	 *       either synchronously or lazily
	 *
	 * FIXME: replace local_flush_tlb_one() with the smp version
	 *        on cpus that belong to a domain
	 */
	local_flush_tlb_one(gpaddr);
    put_page(page);
}

static void adjust_page_count_info(struct page_info* page)
{
    struct domain* d = page_get_owner(page);
    BUG_ON((page->count_info & PGC_count_mask) < 1);
    if (d != NULL) {
        int ret = get_page(page, d);
        BUG_ON(ret == 0);
    } else {
        unsigned long x, nx, y;

        y = page->count_info;
        do {
            x = y;
            nx = x + 1;

            BUG_ON((x >> 32) != 0);
            BUG_ON((nx & PGC_count_mask) != 2);
            y = cmpxchg(&page->count_info, x, nx);
        } while (unlikely(y != x));
        BUG_ON(page_get_owner(page) != NULL);
    }
}

static void domain_put_page(struct domain* d, unsigned long mpaddr,
							pte_t* ptep, pte_t old_pte, int clear_PGC_allocate)
{
    unsigned long mfn = pte_pfn(old_pte);
    struct page_info* page = mfn_to_page(mfn);

    if (pte_present(old_pte)) {
		/*
		 * XXX: _PAGE_PRESENT flag is set only for domains that 
		 *      own the corresponding page. So, if the bit is set, 
		 *      then the freeing domain better be the owner
		 */
        if (page_get_owner(page) == d || page_get_owner(page) == NULL) {
			/*
			 * FIXME: do we currently implement get_gpfn_from_mfn()
			 *        correctly ? Do we populate machine_to_phys_mapping[] ?
			 */
            BUG_ON(get_gpfn_from_mfn(mfn) != (mpaddr >> PAGE_SHIFT));
			set_gpfn_from_mfn(mfn, INVALID_M2P_ENTRY);
        } else
            BUG();

        if (likely(clear_PGC_allocate)) {
            if (!test_and_clear_bit(_PGC_allocated, &page->count_info))
                BUG();
        } else {
            /*
			 * In this case, page reference count mustn't touched.
			 * domain_page_flush_and_put() decrements it, we increment
			 * it in advence. This patch is slow path.
			 *
			 * guest_remove_page(): owner = d, count_info = 1
			 * memory_exchange(): owner = NULL, count_info = 1
			 * XENMEM_add_to_physmap: ower = d, count_info >= 1
			 */
            adjust_page_count_info(page);
        }
    }
    __domain_put_page(d, mpaddr, ptep, old_pte, page);
}

static void assign_domain_page_replace(struct domain *d, unsigned long gpaddr, 
									   unsigned long mfn, unsigned long flags)
{
    struct mm_struct *mm = &d->arch.mm;
    pte_t* pte;
    pte_t old_pte;
    pte_t npte;

    pte = lookup_alloc_domain_pte(d, gpaddr);

    npte = pfn_pte(mfn, vm_get_page_prot(flags)); // build a pte out of mfn

	/*
	 * FIXME: if an old pte is exchanged with a new one
	 *        tlb must be flushed. In MIPS tlbs will
	 *        have to be flushed explicitly. So, read
	 *        the old pte, check if it different from
	 *        the new one and not pte_none(old). So,
	 *        instead of merely using an atomic cmpxchg()
	 *        we might be better off a holding a lock
	 *        on the page table
	 */
    old_pte = ptep_xchg(mm, gpaddr, pte, npte); // xchg it with old pte

	/*
	 * XXX: How do we distinguish between IO pages and
	 *       memory pages ?
	 *
     * if (pte_mem(old_pte)) { 
	 */
    if (!pte_none(old_pte)) { 
        unsigned long old_mfn = pte_pfn(old_pte);

		if (mfn != old_mfn)
            domain_put_page(d, gpaddr, pte, old_pte, 1);
    }
}

/**
 * @gpaddr: In a virtualized world, this would be pseudo physical address
 *          In our case, for the present, it would be host virtual address
 * @mfn: frame: machine page frame
 * @flags: GNTMAP_readonly | GNTMAP_application_map | GNTMAP_contains_pte
 */

int create_grant_host_mapping(unsigned long gpaddr, unsigned long mfn, 
							  unsigned int flags, unsigned int cache_flags)
{
    struct domain* d = current->domain;
    struct page_info* page;
    int ret;

    if ((flags & (GNTMAP_device_map | 
                  GNTMAP_application_map | GNTMAP_contains_pte)) ||
        (cache_flags)) {
        gdprintk(XENLOG_INFO, "%s: flags 0x%x cache_flags 0x%x\n",
                 __func__, flags, cache_flags);
        return GNTST_general_error;
    }
        
    BUG_ON(!mfn_valid(mfn));
    page = mfn_to_page(mfn);
    ret = get_page(page, page_get_owner(page));
    BUG_ON(ret == 0);
    assign_domain_page_replace(d, gpaddr, mfn,
							   ((flags & GNTMAP_readonly) ? VM_READ : VM_READ | VM_WRITE));
        
    return GNTST_okay;
}

int is_iomem_page(unsigned long mfn)
{
	return 0;
//    return (!mfn_valid(mfn) || (page_get_owner(mfn_to_page(mfn)) == dom_io));
}

/*
 * stolen from xen/arch/ia64/xen/mm.c
 */
void share_xen_page_with_guest(struct page_info *page,
							   struct domain *d, int readonly)
{
    if (page_get_owner(page) == d)
        return;

    if (readonly)
        printk("%s:%d readonly is not supported yet\n", __func__, __LINE__);

    /*
	 * alloc_xenheap_pages() doesn't initialize page owner.
	 * 
	 * BUG_ON(page_get_owner(page) != NULL);
	 */
    spin_lock(&d->page_alloc_lock);

#if !defined(__ia64__) && !defined(__mips__)
    /* 
	 * The incremented type count pins as writable or read-only. 
	 */
    page->u.inuse.type_info  = (readonly ? PGT_none : PGT_writable_page);
    page->u.inuse.type_info |= PGT_validated | 1;
#endif

    page_set_owner(page, d);
    wmb(); /* install valid domain ptr before updating refcnt. */
    ASSERT((page->count_info & ~PGC_xen_heap)== 0);

    /* 
	 * Only add to the allocation list if the domain isn't dying. 
	 */
    if (!d->is_dying) {
        page->count_info |= PGC_allocated | 1;
        if ( unlikely(d->xenheap_pages++ == 0) )
            get_knownalive_domain(d);
        page_list_add_tail(page, &d->xenpage_list);
    }

    /*
	 * grant_table_destroy() releases these pages. But it doesn't 
	 * clear their m2p entry. So there might remain stale entries. 
	 * Such a stale entry is cleared here.
	 */
    set_gpfn_from_mfn(page_to_mfn(page), INVALID_M2P_ENTRY);

    spin_unlock(&d->page_alloc_lock);
}

int replace_grant_host_mapping(unsigned long gpaddr,
							   unsigned long mfn, 
							   unsigned long new_gpaddr, 
							   unsigned int flags)
{
	return 0;
}

/*
 * currently, the following is used to map grant table
 * frames into NetOS address space. Linux sets it up
 * in page tables of its own and not the one created by
 * Xen. Might have to change at some point
 */
long  arch_memory_op(int op, XEN_GUEST_HANDLE(void) arg)
{
	struct xen_physmap_mfn_list physmap;
	struct domain *d;
	int i;
	unsigned long gpfn, mfn;
	
	switch (op) {
	case XENMEM_physmap_mfn_list:
		if (copy_from_guest(&physmap, arg, 1) != 0)
			return -EACCES;
		
		if (physmap.domid == DOMID_SELF)
			d = current->domain;
		else
			d = rcu_lock_domain_by_id(physmap.domid);
		if (d == NULL)
			return -EINVAL;
		
		for (i = 0; i < physmap.nr_gpfns; ++i) {
			(void)copy_from_guest_offset(&gpfn, physmap.gpfn_list, 0, 1);
			(void)copy_from_guest_offset(&mfn, physmap.mfn_list, i, 1);
			guest_physmap_add_page(d, gpfn >> PAGE_SHIFT, mfn, 0);
		}

		break;

	default:
		panic("unknown cmd to arch_memory_op\n");
		break;
	}
	return 0;
}

/*
 * heavily depends on the struct page layout. gnttab_transfer() 
 * calls steal_page() with memflags = 0. For grant table transfer, 
 * we must fill the page. memory_exchange() calls steal_page() 
 * with memflags = MEMF_no_refcount. For memory exchange, we 
 * don't have to fill the page because memory_exchange() does it.
 */
int steal_page(struct domain *d, 
			   struct page_info *page, 
			   unsigned int memflags)
{
	return 0;
}

int page_is_ram_type(unsigned long mfn, unsigned long type)
{
	return 0;
}

unsigned long domain_get_maximum_gpfn(struct domain *d)
{
    // return (d->arch.convmem_end - 1) >> PAGE_SHIFT;
	return 0;
} 

void share_xen_page_with_privileged_guests(struct page_info *page, int readonly)
{
    // share_xen_page_with_guest(page, dom_xen, readonly);
}

/*
 * Just as linux builds mem_map[], an array of
 * struct page structures corresponding to each
 * physical page frame, frametable is used for
 * similar purpose by Xen for internal use
 */

/* 
 * FIXME: postpone support to machines with big 
 * holes between physical memorys. Current hack 
 * allows only efi memdesc upto 4G place. 
 * (See efi.c)
 */
#define FT_ALIGN_SIZE	(16UL << 20)
void __init init_frametable(void)
{
	unsigned long pfn;
	unsigned long frame_table_size;

	frame_table_size = max_page * sizeof(struct page_info);
	frame_table_size = (frame_table_size + PAGE_SIZE - 1) & PAGE_MASK;

	/* Request continuous trunk from boot allocator, since HV
	 * address is identity mapped */
	pfn = alloc_boot_pages(
            frame_table_size >> PAGE_SHIFT, FT_ALIGN_SIZE >> PAGE_SHIFT);
	if (pfn == 0)
		panic("Not enough memory for frame table.\n");

	frame_table = __va(pfn << PAGE_SHIFT);
	memset(frame_table, 0, frame_table_size);
}

/*
 * pud, pmd, pte page areas must be visible before population 
 * so that cmpxchg must have release semantics.
 */

/*
 * 'alloc' in the function name indicates that this
 *  version allocates page table pages if they are
 *  needed.
 *
 * FIXME: The following functions have not been tested
 *        with multiple vcpus of a domain. There could be
 *        potential race conditions/synchronization
 *        problems although we use cmpxchg_successful()
 *        to ensure atomicity/mutual exclusion
 */
pte_t* lookup_alloc_domain_pte(struct domain* d, unsigned long mpaddr)
{
    struct mm_struct *mm = &(d->arch.mm);
    volatile pgd_t *pgd;
    volatile pud_t *pud;
    volatile pmd_t *pmd;

    BUG_ON(mm->pgd == NULL);
	
    pgd = pgd_offset(mm, mpaddr);
 again_pgd:
    if (unlikely(pgd_none(*pgd))) {
		/*
		 * we should never be here for 3-level walk since
	     * pud is folded into pgd
		 *
		 * FIXME: replace old_pud = NULL with invalid_pud_table
		 *        if we ever move to a 4-level walk
		 */
        pud_t *old_pud = NULL;
        pud = pud_alloc_one(mm, mpaddr);
		/*
		 * @doc: Xen on multiple cpus can access this structure
		 *       and has to be updated atomically. Linux obtains
		 *       a spinlock while cmpxchg semantics are used here
		 */
		if (unlikely(!cmpxchg_successful(&pgd_val(*pgd), old_pud, pud))) {
            pud_free(pud);
            goto again_pgd;
        }
    }

    pud = pud_offset((pgd_t *)pgd, mpaddr); /* for 2-level walk, pud = pgd */
 again_pud:
    if (unlikely(pud_none(*pud))) {
		/*
		 * we should never be here for 2-level walk
		 */
        pmd_t* old_pmd = invalid_pmd_table;
        pmd = pmd_alloc(mm, mpaddr);
        if (unlikely(!cmpxchg_successful(&pud_val(*pud), old_pmd, pmd))) {
            pmd_free((pmd_t *)pmd);
            goto again_pud;
        }
    }

	/*
	 * pmd_offset() returns pud pointer wrapped in a structure
	 * struct { pud }
	 */
    pmd = pmd_offset((pud_t *)pud, mpaddr);
 again_pmd:
    if (unlikely(pmd_none(*pmd))) {
        pte_t* old_pte = invalid_pte_table;
        pte_t* pte = pte_alloc(mm, mpaddr);
        if (unlikely(!cmpxchg_successful(&pmd_val(*pmd), old_pte, pte))) {
			pte_free(pte);
            goto again_pmd;
        }
    }

    return pte_offset_map(pmd, mpaddr);
}

pte_t* lookup_domain_pte(struct domain* d, unsigned long mpaddr)
{
    struct mm_struct *mm = &(d->arch.mm);
    volatile pgd_t *pgd;
    volatile pud_t *pud;
    volatile pmd_t *pmd;

    BUG_ON(mm->pgd == NULL);
	
    pgd = pgd_offset(mm, mpaddr);
    pud = pud_offset((pgd_t *)pgd, mpaddr); /* for 2-level walk, pud = pgd */
    pmd = pmd_offset((pud_t *)pud, mpaddr);

    return pte_offset_map(pmd, mpaddr);
}

/* 
 * Allocate a new page for domain and map it to the specified
 * metaphysical address. This function is called only while
 * creating new domains. Domain pages allocated during other 
 * times invoke a different set of functions such as
 * guest_physmap_add_page()
 *
 * FIXME: the semantics of the following function is incomplete
 *        as there is no argument to pass permissions of the
 *        for the mapped area.
 */
static struct page_info *
__assign_new_domain_page(struct domain *d, unsigned long mpaddr, pte_t* pte)
{
    struct page_info *p;
    unsigned long maddr;

    BUG_ON(!pte_none(*pte));

    p = alloc_domheap_pages(d, find_first_set_bit(DOM_PAGE_SIZE / PAGE_SIZE), 0);
    if (unlikely(!p)) {
        printk("assign_new_domain_page: Can't alloc!!!! Aaaargh!\n");
        return(p);
    }

    // zero out pages for security reasons
    clear_page(page_to_virt(p));
    maddr = page_to_maddr(p);
	
	/*
	 * Set machine to domain mpaddr mapping
	 */
    set_gpfn_from_mfn((page_to_mfn(p) << PAGE_SHIFT) >> DOM_PAGE_SHIFT, mpaddr >> DOM_PAGE_SHIFT);

	if (pte_present(*pte)) {
        /* 
		 * FIXME: how can this happen ? vhpt is allocated by alloc_domheap_page.  
		 */
        printk("assign_new_domain_page: reassigned vhpt page %lx!!\n", maddr);
    }

    // clear_page() and set_gpfn_from_mfn() become visible before set_pte_rel()
    // because set_pte_rel() has release semantics
	
    set_pte(pte, pfn_pte(maddr >> PAGE_SHIFT,
						 vm_get_page_prot(VM_READ | VM_WRITE | VM_OWNER)));

	smp_mb();

    return p;
}

struct page_info *
assign_new_domain_page(struct domain *d, unsigned long mpaddr)
{
	pte_t *pte;

	pte = lookup_alloc_domain_pte(d, mpaddr);

    if (!pte_none(*pte))
		return mfn_to_page(pte_val(*pte) >> PAGE_SHIFT);

	/*
	 * pte points to an entry in the domains's page table
	 * Do pte entries capture machine address or guest
	 * domains' physical address ? Since this is about
	 * guest domain's address mapping that should be transparent
	 * to domains, we can represent machine addresses in 
	 * domains' page tables. And the page tables themselves
	 * are populated by the virutal addresses native to
	 * Xen, which in MIPS case are unmapped
	 */
    return __assign_new_domain_page(d, mpaddr, pte);
}

int xen_in_range(paddr_t start, paddr_t end)
{
	paddr_t xs = __pa(&_text);
	paddr_t xe = __pa(&_end);

	return (start < xe) && (end > xs);
}

#ifdef CONFIG_XEN_PGWALKER
static int __initdata tlb_config = (ENABLE_ETLB | ENABLE_PGWALKER);
#else
static int __initdata tlb_config = ENABLE_ETLB;
#endif

void tlb_init(void)
{
	write_c0_pagemask(PM_DOM_MASK);

	write_c0_wired(0);

	write_c0_framemask(0);
}

void enable_kernel_segment_access(unsigned int *flags)
{
#define KXBIT 7

	*flags = read_c0_status();

	write_c0_status(*flags | (1 << KXBIT));
}

void mmu_init(void)
{
	unsigned int flags;

	enable_kernel_segment_access(&flags);

	write_c0_config7(PM_DOM_MASK >> (13 + (ffz(PM_DOM_MASK >> 13) / 2)));

	pgwalker_init();

	/* Initialize after pgwalker is configured */
	write_c0_config6(read_c0_config6() | tlb_config);
}

DEFINE_PER_CPU(unsigned long [NR_ADDR_SEGMENTS], pgd_bases);

#ifdef CONFIG_XEN_2LEVEL_PGTABLE
static int pgtable_levels = PGD | PTE;
#else
static int pgtable_levels = PGD | PMD | PTE;
#endif

void pgwalker_init(void)
{
	unsigned int value, i;

	if (tlb_config & ENABLE_PGWALKER) {

		/* 
		 * hardware page levels information:
		 * 
		 * [15:8] no of top-most bits of vaddr used to form
		 *        an index into the pgdirs table
		 * [ 7:4] shift amount by which pfn (page frame number)
		 *        needs to be left shifted for populating the
		 *        entrylo0 and entrylo1 registers
		 * [ 3:0] page table levels used. 32-bit kernels use 
		 *        pgd and pte levels, while 64-bits kernels
		 *        use pgd, pmd, and pte
		 */
		value  = ((ffs(NR_ADDR_SEGMENTS) - 1) & 0xff) << 8;
		value |= ENTRYLO_PFN_SHIFT << 4;
		value |= pgtable_levels;
		pgw_register_write_w(PGW_MMU_INFO, value);

		for (i = 0; i < NR_ADDR_SEGMENTS; i++)
			__get_cpu_var(pgd_bases)[i] = (unsigned long)invalid_pg_dir;

#ifdef CONFIG_XEN_64BIT
		pgw_register_write_d(PGW_PGD_BASES, (unsigned long)&(__get_cpu_var(pgd_bases)[0]));
#else
		pgw_register_write_w(PGW_PGD_BASES, (unsigned int)&(__get_cpu_var(pgd_bases)[0]));
#endif

		/* PGD shift and mask information */
		pgw_register_write_w(PGW_PGD_SHIFT, PGDIR_SHIFT - PTRLOG);
		pgw_register_write_w(PGW_PGD_MASK, (PTRS_PER_PGD - 1) << PTRLOG);

#ifndef CONFIG_XEN_2LEVEL_PGTABLE
		/*
		 * MIPS Linux currently does not use 4-level page tables
		 * and hence it is not necessary to fill in pud information
		 *
		 * So, just fill in the PMD shift and mask information 
		 */
		pgw_register_write_w(PGW_PMD_SHIFT, PMD_SHIFT - PTRLOG);
		pgw_register_write_w(PGW_PMD_MASK, (PTRS_PER_PMD - 1) << PTRLOG);
#endif

		/* PTE shift and mask */
		pgw_register_write_w(PGW_PTE_SHIFT, DOM_PAGE_SHIFT - PTRLOG);
		pgw_register_write_w(PGW_PTE_MASK, (PTRS_PER_PTE - 1) << PTRLOG);

		put_cpu_var(pgd_bases);

		// dump_pgwalker_config();
		printk("Initialized Page Walker on cpu %d\n", processor_id());
	}
}

void dump_pgwalker_config(void)
{
    pgw_print_w(PGW_MMU_INFO);
    pgw_print_w(PGW_PGD_SHIFT);
    pgw_print_w(PGW_PGD_MASK);
    pgw_print_w(PGW_PMD_SHIFT);
    pgw_print_w(PGW_PMD_MASK);
    pgw_print_w(PGW_PTE_SHIFT);
    pgw_print_w(PGW_PTE_MASK);
}

void __guest_physmap_add_page(struct domain *d, unsigned long gpfn,
									 unsigned long mfn)
{
    set_gpfn_from_mfn(mfn, gpfn);
    smp_mb();
	/*
	 * XXX: The pages being mapped into the domain must belong to the 
	 *      domain which is why we set the VM_OWNER flag. Alternatively, 
	 *      pages belonging to another domaint too could be mapped into 
	 *      a domain. That mapping should take another code path. 
	 *      Verify this.
	 */

	/*
	 * FIXME: replace local_flush_tlb_one() with smp version
	 *        shut off the page walker before smp_mb() above
	 */
	local_flush_tlb_one(gpfn << PAGE_SHIFT);
    assign_domain_page_replace(d, gpfn << PAGE_SHIFT, mfn,
                               VM_READ | VM_WRITE | VM_OWNER);

	/*
	 * XXX: need more functionality here if iommu is enabled
	 */
}

/*
 * In grant table context, the following functions is called only
 * while transferring pages (change of ownership) and not while
 * mapping shared pages. Hence 'ASSIGN_pgc_allocated' flag is set
 *
 * There are other memory operations that a guest can perform
 * outside the grant_table context ; such as mapping more physical
 * memory into its address space. guest_physmap_add_page() could
 * be called in those contexts too
 */ 

int guest_physmap_add_page(struct domain *d, unsigned long gpfn,
						   unsigned long mfn, unsigned int page_order)
{
    unsigned long i;

    for (i = 0; i < (1UL << page_order); i++) {
        BUG_ON(!mfn_valid(mfn));
        BUG_ON((mfn_to_page(mfn)->count_info & ~PGC_xen_heap) !=
               (PGC_allocated | 1));
        __guest_physmap_add_page(d, gpfn, mfn);
        mfn++;
        gpfn++;
    }

	return 0;
}

static inline void setup_protection_map(void)
{
	protection_map[0] = PAGE_NONE;
	protection_map[1] = PAGE_READONLY;
	protection_map[2] = PAGE_RW;
	protection_map[3] = PAGE_RW;
	protection_map[4] = PAGE_READONLY;
	protection_map[5] = PAGE_READONLY;
	protection_map[6] = PAGE_RW;
	protection_map[7] = PAGE_RW;

	protection_map[8] = PAGE_NONE_OWNER;
	protection_map[9] = PAGE_READONLY_OWNER;
	protection_map[10] = PAGE_RW_OWNER;
	protection_map[11] = PAGE_RW_OWNER;
	protection_map[12] = PAGE_READONLY_OWNER;
	protection_map[13] = PAGE_READONLY_OWNER;
	protection_map[14] = PAGE_RW_OWNER;
	protection_map[15] = PAGE_RW_OWNER;
}

void mm_init(void)
{
	setup_protection_map();

	pgd_init((unsigned long)invalid_pg_dir);
	pmd_init((unsigned long)invalid_pmd_table);
	pte_init((unsigned long)invalid_pte_table);
}

static void
mm_teardown_pmd(struct domain* d, pmd_t* pmd,
                    unsigned long offset)
{
	pte_free(pte_offset(pmd, offset));
}

static void
mm_teardown_pud(struct domain* d, pud_t *pud,
                    unsigned long offset)
{
    unsigned long i;
    pmd_t *pmd = pmd_offset(pud, offset);

    for (i = 0; i < PTRS_PER_PMD; i++, pmd++) {
        if (!pmd_present(*pmd))
            continue;
        mm_teardown_pmd(d, pmd, offset + (i << PMD_SHIFT));
    }
    pmd_free(pmd_offset(pud, offset));
}

static void
mm_teardown_pgd(struct domain* d, pgd_t *pgd,
                    unsigned long offset)
{
    unsigned long i;
    pud_t *pud = pud_offset(pgd, offset);

    for (i = 0; i < PTRS_PER_PUD; i++, pud++) {
        if (!pud_present(*pud))
            continue;
        mm_teardown_pud(d, pud, offset + (i << PUD_SHIFT));
    }
    pud_free(pud_offset(pgd, offset));
}

void mm_teardown(struct domain *d)
{
	struct mm_struct* mm = &d->arch.mm;
    unsigned long i;
    pgd_t* pgd;

    BUG_ON(mm->pgd == NULL);
    pgd = pgd_offset(mm, 0UL);
    for (i = 0; i < PTRS_PER_PGD; i++, pgd++) {
        if (!pgd_present(*pgd))
            continue;
        mm_teardown_pgd(d, pgd, i << PGDIR_SHIFT);
    }
    pgd_free(mm->pgd);
    mm->pgd = NULL;
}
