#include <xen/config.h>
#include <xen/threads.h>
#include <xen/sched.h>
#include <xen/types.h>
#include <xen/cache.h>
#include <xen/cpumask.h>
#include <xen/smp.h>
#include <xen/mm.h>
#include <xen/debug.h>
#include <xen/domain.h>

#include <asm/system.h>
#include <asm/processor.h>
#include <asm/cpu.h>
#include <asm/tlbflush.h>
#include <asm/current.h>
#include <asm/spinlock.h>
#include <asm/hardirq.h>
#include <asm/addrspace.h>
#include <asm/netl/asmmacro.h>
#include <asm/netl/pic.h>
#include <asm/traps.h>
#include <asm/delay.h>

cpumask_t cpu_sibling_map[NR_CPUS] __cacheline_aligned;

cpumask_t cpu_online_map;
cpumask_t cpu_present_map __read_mostly;
cpumask_t phys_cpu_present_map;
cpumask_t cpu_possible_map;
cpumask_t cpu_callin_map;

static cpumask_t cpu_cond_variable;

cpumask_t cpu_core_map[NR_CPUS];

int __cpu_number_map[NR_CPUS];      /* Map physical to logical */
int __cpu_logical_map[NR_CPUS];     /* Map logical to physical */

#ifdef CONFIG_XEN_SMP

extern void tail_jump(unsigned long, unsigned long, unsigned long);
extern void startup_cpu_idle_loop(void);

#define PER_CPU_THREAD_SIZE    (THREAD_SIZE >> 2)
#define TOTAL_THREAD_SIZE      (PER_CPU_THREAD_SIZE * (NR_CPUS - 1))

struct stack {
	unsigned long stack[TOTAL_THREAD_SIZE/sizeof(long)];
};

struct stack secondary_stack __attribute__((__section__(".data.init_task"), __aligned__(THREAD_SIZE)));

static DEFINE_SPINLOCK(smp_lock);
static DEFINE_SPINLOCK(cpumask_lock);

/*
 * The following two API's must be used when attempting
 * to serialize the updates to cpu_online_mask, cpu_present_mask.
 */
void cpu_maps_update_begin(void)
{
	spin_lock(&cpumask_lock);
}

void cpu_maps_update_done(void)
{
	spin_unlock(&cpumask_lock);
}

static void flush_tlb_all_ipi(void *info)
{
	local_flush_tlb_all(); /* defined in tlb-r4k.c */
}

static void flush_tlb_one_ipi(void *info)
{
	unsigned long vaddr = (unsigned long) info;

	local_flush_tlb_one(vaddr);
}

static struct call_data_struct *call_data;

/*
 * Run a function on all other CPUs.
 *  <func>      The function to run. This must be fast and non-blocking.
 *  <info>      An arbitrary pointer to pass to the function.
 *  <retry>     If true, keep retrying until ready.
 *  <wait>      If true, wait until function has completed on other CPUs.
 *  [RETURNS]   0 on success, else a negative status code.
 *
 * Does not return until remote CPUs are nearly ready to execute <func>
 * or are or have executed.
 *
 * You must not call this function with disabled interrupts or from a
 * hardware interrupt handler or from a bottom half handler:
 *
 * CPU A                               CPU B
 * Disable interrupts
 *                                     smp_call_function()
 *                                     Take call_lock
 *                                     Send IPIs
 *                                     Wait for all cpus to acknowledge IPI
 *                                     CPU A has not responded, spin waiting
 *                                     for cpu A to respond, holding call_lock
 * smp_call_function()
 * Spin waiting for call_lock
 * Deadlock                            Deadlock
 */
int smp_call_function(void (*func) (void *info), void *info, int retry,
								int wait)
{
	struct call_data_struct data;
	int i, cpus = num_online_cpus() - 1;
	int cpu = smp_processor_id();

	/*
	 * Can die spectacularly if this CPU isn't yet marked online
	 */
	BUG_ON(!cpu_online(cpu));

	if (!cpus)
		return 0;

	/* Can deadlock when called with interrupts disabled */
	WARN_ON(irqs_disabled());

	data.func = func;
	data.info = info;
	atomic_set(&data.started, 0);
	data.wait = wait;
	if (wait)
		atomic_set(&data.finished, 0);

	spin_lock(&smp_lock);
	call_data = &data;
	smp_mb();

	/* Send a message to all other CPUs and wait for them to respond */
	for_each_online_cpu(i)
		if (i != cpu)
			core_send_ipi(i, IPI_CALL_FUNCTION);

	/* Wait for response */
	/* FIXME: lock-up detection, backtrace on lock-up */
	while (atomic_read(&data.started) != cpus)
		barrier();

	if (wait)
		while (atomic_read(&data.finished) != cpus)
			barrier();
	call_data = NULL;
	spin_unlock(&smp_lock);

	return 0;
}

void flush_tlb_all(void)
{
	on_each_cpu(flush_tlb_all_ipi, NULL, 1, 1);
}

void flush_tlb_one(unsigned long vaddr)
{
	smp_call_function(flush_tlb_one_ipi, (void *)vaddr, 1, 1);

	flush_tlb_one_ipi((void *)vaddr);
}

void flush_tlb_mask(cpumask_t mask)
{

}

void smp_send_event_check_mask(cpumask_t mask)
{
    int cpu;

    /*  Not for me.  */
    cpu_clear(smp_processor_id(), mask);
    if (cpus_empty(mask))
        return;

    for (cpu = 0; cpu < NR_CPUS; ++cpu)
        if (cpu_isset(cpu, mask))
			core_send_ipi(cpu, IPI_EVENT_CHECK_VECTOR);
}

struct call_data_struct call_data_struct;

int on_selected_cpus(cpumask_t selected, 
					 void (*func) (void *info), void *info,
					 int retry, int wait)
{
    unsigned int cpu, nr_cpus = cpus_weight(selected);

    ASSERT(local_irq_is_enabled());

    if (!nr_cpus)
        return 0;

   spin_lock(&smp_lock);

    call_data_struct.func = func;
    call_data_struct.info = info;
    call_data_struct.wait = wait;
    atomic_set(&call_data_struct.started, 0); 
    atomic_set(&call_data_struct.finished, 0);

    call_data = &call_data_struct;

    wmb();

    for_each_cpu_mask(cpu, selected) {
        core_send_ipi(cpu, IPI_CALL_FUNCTION);
    }

    while (atomic_read(wait ? &call_data_struct.finished : &call_data_struct.started) != nr_cpus)
        cpu_relax();

    spin_unlock(&smp_lock);

    return 0;
}

void smp_call_function_interrupt(void)
{
	void (*func) (void *info) = call_data->func;
	void *info = call_data->info;
	int wait = call_data->wait;

	/*
	 * Notify initiating CPU that I've grabbed the data and am
	 * about to execute the function.
	 */
	smp_mb();

	atomic_inc(&call_data->started);

	/*
	 * At this point the info structure may be out of scope unless wait==1.
	 */
	// irq_enter();
	(*func)(info);
	// irq_exit();

	if (wait) {
		smp_mb();
		atomic_inc(&call_data->finished);
	}
}

#define SET_IPI_VECTOR(x, y) x |= y

void core_send_ipi(int logical_cpu, unsigned int action)
{
	int cpu = cpu_logical_map(logical_cpu);
	__u32 ipi = 0;

	if (action == IPI_CALL_FUNCTION) {
		SET_IPI_VECTOR(ipi, IRQ_IPI_SMP_FUNCTION);
	} else if (action == SMP_RESCHEDULE_YOURSELF) {
		SET_IPI_VECTOR(ipi, IRQ_IPI_SMP_RESCHEDULE);
	} else if (action == IPI_EVENT_CHECK_VECTOR) {
		SET_IPI_VECTOR(ipi, IRQ_IPI_SMP_RESCHEDULE);
	}
	else
		BUG();
#if defined(XLP_SIM)
	pic_send_ipi(0, (ipi & 0x3f), 0, cpu);
#else
	pic_send_ipi(ipi);
#endif
}

void handle_ipi(int irq, void *arg, struct pt_regs *regs)
{
	if (irq == IRQ_IPI_SMP_FUNCTION) {
		smp_call_function_interrupt();
	}
	else {
		/*
		 * case for reschedule. nothing to do.
		 * do_softirq() kicks in when irq_exit()
		 * is called after the call to handle_ipi()
		 */
	}
}

void init_secondary(void)
{
	/* Set up kseg0 to be cachable coherent */
	change_c0_config(CONF_CM_CMASK, CONF_CM_DEFAULT);

	//init_xen_time();

	init_IRQ();
}

void smp_finish(void)
{
	// msgring_cpu_init();
}

extern void secondary_cpu_probe(void);

extern void	build_dom1(void);
extern int dom1_cpu;

asmlinkage void resume_secondary_cpus(void)
{
	unsigned int cpu;

	cpu = smp_processor_id();

	cpu_probe();

	trap_init();

	init_secondary();

	local_irq_enable();

	mmu_init();

	tlb_init();

	cpu_probe();

	smp_finish();

	cpu_set(cpu, cpu_callin_map);

	if (cpu == dom1_cpu) {
		cpu_clear(cpu_number_map(cpu), cpu_online_map);
		build_dom1();
	}

	startup_cpu_idle_loop();
}

void start_secondary(void)
{
	struct vcpu *task;

	write_c0_ebase((uint32_t)ebase);
	
	write_c0_context(processor_id() << (CP0_CONTEXT_PTEBASE_SHIFT + LONGLOG));

	cpu_maps_update_begin();

	cpu_set(processor_id(), phys_cpu_present_map);

	cpu_maps_update_done();

	while (!cpu_isset(processor_id(), cpu_cond_variable))
		barrier();


	task = idle_vcpu[__cpu_number_map[processor_id()]];

	tail_jump((unsigned long)resume_secondary_cpus, (unsigned long)__KSTK_TOS(task), (unsigned long)task_thread_info(task));
}

static int cpu_up(unsigned int cpu)
{
	if (!idle_vcpu[cpu]) {
		idle_vcpu[cpu] = alloc_vcpu(idle_vcpu[0]->domain, cpu, cpu);
		if (idle_vcpu[cpu] == NULL)
			panic("panic...unable to allocate vpuc\n");
	} 
	else {
		/*
		 * we should not be here
		 */
		printk(" ...........\n");
		cpu_halt();
	}

	cpu_set(__cpu_logical_map[cpu], cpu_cond_variable);

	/*
	 * Trust is futile. We should really have timeouts ...
	 */
	while (!cpu_isset(cpu, cpu_callin_map))
		barrier();

	cpu_set(cpu, cpu_online_map);

	/*
	 * XXX Set the cpu_sibling_map[] here ?
	 */
	cpu_set(0, cpu_sibling_map[cpu]);

	return 0;
}

extern char __per_cpu_start[], __per_cpu_data_end[], __per_cpu_end[];

static void __init percpu_init_areas(void)
{
    unsigned int i, data_size = __per_cpu_data_end - __per_cpu_start;

    BUG_ON(data_size > PERCPU_SIZE);
    /* 
	 * Initialise per-cpu data area for all possible secondary CPUs. 
	 */
    for ( i = 1; (i < NR_CPUS) && cpu_possible(i); i++ )
        memcpy(__per_cpu_start + (i << PERCPU_SHIFT),
               __per_cpu_start,
               data_size);
}

/* 
 * Called by boot processor to activate the rest 
 */
void smp_init(void)
{
	int ncpu, cpu;

	/*
	 * Online cpus are determined during bootup by sending a
	 * wakeup signal to each of "NR_CPUS - 1" cpus. Online
	 * cpus then set the corresponding bit in phys_cpu_present_map
	 */
	cpus_clear(cpu_possible_map);

	for (cpu = ncpu = 0; cpu < NR_CPUS; ++cpu) {
		if (cpu_isset(cpu, phys_cpu_present_map)) {
			__cpu_number_map[cpu] = ncpu;
			__cpu_logical_map[ncpu] = cpu;
			cpu_set(ncpu, cpu_possible_map);
			++ncpu;
		}
	}

	cpu_present_map = cpu_possible_map;

	percpu_init_areas();
}

void boot_secondary(void)
{
	int cpu;

	/* 
	 * cpus are woken up based on their logical id, 
	 * as opposed to physical id
	 */
	for_each_present_cpu(cpu) {
		if (!cpu_online(cpu))
			cpu_up(cpu);
	}

	printk(KERN_INFO "Brought up %ld CPUs\n", (long)num_online_cpus());
}

int init_ipi(void)
{
	int retval;

	retval = request_irq_vector(IRQ_IPI_SMP_FUNCTION, handle_ipi, IRQ_PER_CPU, "timer", NULL);
	if ((retval != 0) && (retval != -EBUSY)) {
		printk("Unable to register handler for irq %d\n", IRQ_IPI_SMP_FUNCTION);
		goto exit;
	}
	
	retval = request_irq_vector(IRQ_IPI_SMP_RESCHEDULE, handle_ipi, IRQ_PER_CPU, "timer", NULL);
	if ((retval != 0) && (retval != -EBUSY)) {
		printk("Unable to register handler for irq %d\n", IRQ_IPI_SMP_RESCHEDULE);
		goto exit;
	}
	
exit:
	return retval;
}

#else /* ! CONFIG_XEN_SMP */

int smp_call_function (void (*func) (void *info), void *info, int retry, int wait)
{
	return 0;
}

void smp_send_event_check_mask(cpumask_t mask)
{

}

int on_selected_cpus(cpumask_t selected, 
					 void (*func) (void *info), void *info,
					 int retry, int wait)
{
	(*func)(info);

	return 0;
}

void smp_init(void) { }

#endif
